/* Generated by libdrgn/build-aux/gen_arch_inc_strswitch.py. */

enum {
	DRGN_REGISTER_NUMBER__lr,
	DRGN_REGISTER_NUMBER__r0,
	DRGN_REGISTER_NUMBER__r1,
	DRGN_REGISTER_NUMBER__r2,
	DRGN_REGISTER_NUMBER__r3,
	DRGN_REGISTER_NUMBER__r4,
	DRGN_REGISTER_NUMBER__r5,
	DRGN_REGISTER_NUMBER__r6,
	DRGN_REGISTER_NUMBER__r7,
	DRGN_REGISTER_NUMBER__r8,
	DRGN_REGISTER_NUMBER__r9,
	DRGN_REGISTER_NUMBER__r10,
	DRGN_REGISTER_NUMBER__r11,
	DRGN_REGISTER_NUMBER__r12,
	DRGN_REGISTER_NUMBER__r13,
	DRGN_REGISTER_NUMBER__r14,
	DRGN_REGISTER_NUMBER__r15,
	DRGN_REGISTER_NUMBER__r16,
	DRGN_REGISTER_NUMBER__r17,
	DRGN_REGISTER_NUMBER__r18,
	DRGN_REGISTER_NUMBER__r19,
	DRGN_REGISTER_NUMBER__r20,
	DRGN_REGISTER_NUMBER__r21,
	DRGN_REGISTER_NUMBER__r22,
	DRGN_REGISTER_NUMBER__r23,
	DRGN_REGISTER_NUMBER__r24,
	DRGN_REGISTER_NUMBER__r25,
	DRGN_REGISTER_NUMBER__r26,
	DRGN_REGISTER_NUMBER__r27,
	DRGN_REGISTER_NUMBER__r28,
	DRGN_REGISTER_NUMBER__r29,
	DRGN_REGISTER_NUMBER__r30,
	DRGN_REGISTER_NUMBER__r31,
	DRGN_REGISTER_NUMBER__cr0,
	DRGN_REGISTER_NUMBER__cr1,
	DRGN_REGISTER_NUMBER__cr2,
	DRGN_REGISTER_NUMBER__cr3,
	DRGN_REGISTER_NUMBER__cr4,
	DRGN_REGISTER_NUMBER__cr5,
	DRGN_REGISTER_NUMBER__cr6,
	DRGN_REGISTER_NUMBER__cr7,
};

struct drgn_arch_register_layout {
	char lr[8];
	char r0[8];
	char r1[8];
	char r2[8];
	char r3[8];
	char r4[8];
	char r5[8];
	char r6[8];
	char r7[8];
	char r8[8];
	char r9[8];
	char r10[8];
	char r11[8];
	char r12[8];
	char r13[8];
	char r14[8];
	char r15[8];
	char r16[8];
	char r17[8];
	char r18[8];
	char r19[8];
	char r20[8];
	char r21[8];
	char r22[8];
	char r23[8];
	char r24[8];
	char r25[8];
	char r26[8];
	char r27[8];
	char r28[8];
	char r29[8];
	char r30[8];
	char r31[8];
	char cr0[8];
	char cr1[8];
	char cr2[8];
	char cr3[8];
	char cr4[8];
	char cr5[8];
	char cr6[8];
	char cr7[8];
};

static const struct drgn_register_layout register_layout[] = {
	{ offsetof(struct drgn_arch_register_layout, lr), 8 },
	{ offsetof(struct drgn_arch_register_layout, r0), 8 },
	{ offsetof(struct drgn_arch_register_layout, r1), 8 },
	{ offsetof(struct drgn_arch_register_layout, r2), 8 },
	{ offsetof(struct drgn_arch_register_layout, r3), 8 },
	{ offsetof(struct drgn_arch_register_layout, r4), 8 },
	{ offsetof(struct drgn_arch_register_layout, r5), 8 },
	{ offsetof(struct drgn_arch_register_layout, r6), 8 },
	{ offsetof(struct drgn_arch_register_layout, r7), 8 },
	{ offsetof(struct drgn_arch_register_layout, r8), 8 },
	{ offsetof(struct drgn_arch_register_layout, r9), 8 },
	{ offsetof(struct drgn_arch_register_layout, r10), 8 },
	{ offsetof(struct drgn_arch_register_layout, r11), 8 },
	{ offsetof(struct drgn_arch_register_layout, r12), 8 },
	{ offsetof(struct drgn_arch_register_layout, r13), 8 },
	{ offsetof(struct drgn_arch_register_layout, r14), 8 },
	{ offsetof(struct drgn_arch_register_layout, r15), 8 },
	{ offsetof(struct drgn_arch_register_layout, r16), 8 },
	{ offsetof(struct drgn_arch_register_layout, r17), 8 },
	{ offsetof(struct drgn_arch_register_layout, r18), 8 },
	{ offsetof(struct drgn_arch_register_layout, r19), 8 },
	{ offsetof(struct drgn_arch_register_layout, r20), 8 },
	{ offsetof(struct drgn_arch_register_layout, r21), 8 },
	{ offsetof(struct drgn_arch_register_layout, r22), 8 },
	{ offsetof(struct drgn_arch_register_layout, r23), 8 },
	{ offsetof(struct drgn_arch_register_layout, r24), 8 },
	{ offsetof(struct drgn_arch_register_layout, r25), 8 },
	{ offsetof(struct drgn_arch_register_layout, r26), 8 },
	{ offsetof(struct drgn_arch_register_layout, r27), 8 },
	{ offsetof(struct drgn_arch_register_layout, r28), 8 },
	{ offsetof(struct drgn_arch_register_layout, r29), 8 },
	{ offsetof(struct drgn_arch_register_layout, r30), 8 },
	{ offsetof(struct drgn_arch_register_layout, r31), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr0), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr1), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr2), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr3), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr4), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr5), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr6), 8 },
	{ offsetof(struct drgn_arch_register_layout, cr7), 8 },
};

static drgn_register_number dwarf_regno_to_internal(uint64_t dwarf_regno)
{
	switch (dwarf_regno) {
	case 65:
		return DRGN_REGISTER_NUMBER__lr;
	case 0:
		return DRGN_REGISTER_NUMBER__r0;
	case 1:
		return DRGN_REGISTER_NUMBER__r1;
	case 2:
		return DRGN_REGISTER_NUMBER__r2;
	case 3:
		return DRGN_REGISTER_NUMBER__r3;
	case 4:
		return DRGN_REGISTER_NUMBER__r4;
	case 5:
		return DRGN_REGISTER_NUMBER__r5;
	case 6:
		return DRGN_REGISTER_NUMBER__r6;
	case 7:
		return DRGN_REGISTER_NUMBER__r7;
	case 8:
		return DRGN_REGISTER_NUMBER__r8;
	case 9:
		return DRGN_REGISTER_NUMBER__r9;
	case 10:
		return DRGN_REGISTER_NUMBER__r10;
	case 11:
		return DRGN_REGISTER_NUMBER__r11;
	case 12:
		return DRGN_REGISTER_NUMBER__r12;
	case 13:
		return DRGN_REGISTER_NUMBER__r13;
	case 14:
		return DRGN_REGISTER_NUMBER__r14;
	case 15:
		return DRGN_REGISTER_NUMBER__r15;
	case 16:
		return DRGN_REGISTER_NUMBER__r16;
	case 17:
		return DRGN_REGISTER_NUMBER__r17;
	case 18:
		return DRGN_REGISTER_NUMBER__r18;
	case 19:
		return DRGN_REGISTER_NUMBER__r19;
	case 20:
		return DRGN_REGISTER_NUMBER__r20;
	case 21:
		return DRGN_REGISTER_NUMBER__r21;
	case 22:
		return DRGN_REGISTER_NUMBER__r22;
	case 23:
		return DRGN_REGISTER_NUMBER__r23;
	case 24:
		return DRGN_REGISTER_NUMBER__r24;
	case 25:
		return DRGN_REGISTER_NUMBER__r25;
	case 26:
		return DRGN_REGISTER_NUMBER__r26;
	case 27:
		return DRGN_REGISTER_NUMBER__r27;
	case 28:
		return DRGN_REGISTER_NUMBER__r28;
	case 29:
		return DRGN_REGISTER_NUMBER__r29;
	case 30:
		return DRGN_REGISTER_NUMBER__r30;
	case 31:
		return DRGN_REGISTER_NUMBER__r31;
	case 68:
		return DRGN_REGISTER_NUMBER__cr0;
	case 69:
		return DRGN_REGISTER_NUMBER__cr1;
	case 70:
		return DRGN_REGISTER_NUMBER__cr2;
	case 71:
		return DRGN_REGISTER_NUMBER__cr3;
	case 72:
		return DRGN_REGISTER_NUMBER__cr4;
	case 73:
		return DRGN_REGISTER_NUMBER__cr5;
	case 74:
		return DRGN_REGISTER_NUMBER__cr6;
	case 75:
		return DRGN_REGISTER_NUMBER__cr7;
	default:
		return DRGN_REGISTER_NUMBER_UNKNOWN;
	}
}

static const struct drgn_register registers[] = {
	{
		.names = (const char * const []){
			"r0",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r0,
	},
	{
		.names = (const char * const []){
			"r1",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r1,
	},
	{
		.names = (const char * const []){
			"r2",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r2,
	},
	{
		.names = (const char * const []){
			"r3",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r3,
	},
	{
		.names = (const char * const []){
			"r4",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r4,
	},
	{
		.names = (const char * const []){
			"r5",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r5,
	},
	{
		.names = (const char * const []){
			"r6",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r6,
	},
	{
		.names = (const char * const []){
			"r7",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r7,
	},
	{
		.names = (const char * const []){
			"r8",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r8,
	},
	{
		.names = (const char * const []){
			"r9",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r9,
	},
	{
		.names = (const char * const []){
			"r10",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r10,
	},
	{
		.names = (const char * const []){
			"r11",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r11,
	},
	{
		.names = (const char * const []){
			"r12",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r12,
	},
	{
		.names = (const char * const []){
			"r13",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r13,
	},
	{
		.names = (const char * const []){
			"r14",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r14,
	},
	{
		.names = (const char * const []){
			"r15",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r15,
	},
	{
		.names = (const char * const []){
			"r16",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r16,
	},
	{
		.names = (const char * const []){
			"r17",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r17,
	},
	{
		.names = (const char * const []){
			"r18",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r18,
	},
	{
		.names = (const char * const []){
			"r19",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r19,
	},
	{
		.names = (const char * const []){
			"r20",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r20,
	},
	{
		.names = (const char * const []){
			"r21",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r21,
	},
	{
		.names = (const char * const []){
			"r22",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r22,
	},
	{
		.names = (const char * const []){
			"r23",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r23,
	},
	{
		.names = (const char * const []){
			"r24",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r24,
	},
	{
		.names = (const char * const []){
			"r25",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r25,
	},
	{
		.names = (const char * const []){
			"r26",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r26,
	},
	{
		.names = (const char * const []){
			"r27",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r27,
	},
	{
		.names = (const char * const []){
			"r28",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r28,
	},
	{
		.names = (const char * const []){
			"r29",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r29,
	},
	{
		.names = (const char * const []){
			"r30",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r30,
	},
	{
		.names = (const char * const []){
			"r31",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__r31,
	},
	{
		.names = (const char * const []){
			"lr",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__lr,
	},
	{
		.names = (const char * const []){
			"cr0",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr0,
	},
	{
		.names = (const char * const []){
			"cr1",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr1,
	},
	{
		.names = (const char * const []){
			"cr2",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr2,
	},
	{
		.names = (const char * const []){
			"cr3",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr3,
	},
	{
		.names = (const char * const []){
			"cr4",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr4,
	},
	{
		.names = (const char * const []){
			"cr5",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr5,
	},
	{
		.names = (const char * const []){
			"cr6",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr6,
	},
	{
		.names = (const char * const []){
			"cr7",
		},
		.num_names = 1,
		.regno = DRGN_REGISTER_NUMBER__cr7,
	},
};

static const struct drgn_register *register_by_name(const char *name)
{
	/* Generated by libdrgn/build-aux/gen_strswitch.py. */
	switch (1) {
	default: {
		#define strswitch0_args(str) \
			const char *strswitch0_str = (str);
		strswitch0_args(
	            name
		)
		#undef strswitch0_args
		if (strswitch0_str[0] == 'c') {
			if (strswitch0_str[1] == 'r') {
				if (strswitch0_str[2] == '0') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case33:
		return &registers[33];
						goto strswitch0_case34;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '1') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case34:
		return &registers[34];
						goto strswitch0_case35;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '2') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case35:
		return &registers[35];
						goto strswitch0_case36;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '3') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case36:
		return &registers[36];
						goto strswitch0_case37;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '4') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case37:
		return &registers[37];
						goto strswitch0_case38;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '5') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case38:
		return &registers[38];
						goto strswitch0_case39;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '6') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case39:
		return &registers[39];
						goto strswitch0_case40;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '7') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case40:
		return &registers[40];
						goto strswitch0_default;
					} else {
						goto strswitch0_default;
					}
				} else {
					goto strswitch0_default;
				}
			} else {
				goto strswitch0_default;
			}
		} else if (strswitch0_str[0] == 'l') {
			if (strswitch0_str[1] == 'r' && strswitch0_str[2] == '\0') {
				strswitch0_case32:
		return &registers[32];
				goto strswitch0_case33;
			} else {
				goto strswitch0_default;
			}
		} else if (strswitch0_str[0] == 'r') {
			if (strswitch0_str[1] == '0') {
				if (strswitch0_str[2] == '\0') {
		return &registers[0];
					goto strswitch0_case1;
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '1') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case1:
		return &registers[1];
					goto strswitch0_case2;
				} else if (strswitch0_str[2] == '0') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case10:
		return &registers[10];
						goto strswitch0_case11;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '1') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case11:
		return &registers[11];
						goto strswitch0_case12;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '2') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case12:
		return &registers[12];
						goto strswitch0_case13;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '3') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case13:
		return &registers[13];
						goto strswitch0_case14;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '4') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case14:
		return &registers[14];
						goto strswitch0_case15;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '5') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case15:
		return &registers[15];
						goto strswitch0_case16;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '6') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case16:
		return &registers[16];
						goto strswitch0_case17;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '7') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case17:
		return &registers[17];
						goto strswitch0_case18;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '8') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case18:
		return &registers[18];
						goto strswitch0_case19;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '9') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case19:
		return &registers[19];
						goto strswitch0_case20;
					} else {
						goto strswitch0_default;
					}
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '2') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case2:
		return &registers[2];
					goto strswitch0_case3;
				} else if (strswitch0_str[2] == '0') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case20:
		return &registers[20];
						goto strswitch0_case21;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '1') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case21:
		return &registers[21];
						goto strswitch0_case22;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '2') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case22:
		return &registers[22];
						goto strswitch0_case23;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '3') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case23:
		return &registers[23];
						goto strswitch0_case24;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '4') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case24:
		return &registers[24];
						goto strswitch0_case25;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '5') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case25:
		return &registers[25];
						goto strswitch0_case26;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '6') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case26:
		return &registers[26];
						goto strswitch0_case27;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '7') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case27:
		return &registers[27];
						goto strswitch0_case28;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '8') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case28:
		return &registers[28];
						goto strswitch0_case29;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '9') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case29:
		return &registers[29];
						goto strswitch0_case30;
					} else {
						goto strswitch0_default;
					}
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '3') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case3:
		return &registers[3];
					goto strswitch0_case4;
				} else if (strswitch0_str[2] == '0') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case30:
		return &registers[30];
						goto strswitch0_case31;
					} else {
						goto strswitch0_default;
					}
				} else if (strswitch0_str[2] == '1') {
					if (strswitch0_str[3] == '\0') {
						strswitch0_case31:
		return &registers[31];
						goto strswitch0_case32;
					} else {
						goto strswitch0_default;
					}
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '4') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case4:
		return &registers[4];
					goto strswitch0_case5;
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '5') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case5:
		return &registers[5];
					goto strswitch0_case6;
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '6') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case6:
		return &registers[6];
					goto strswitch0_case7;
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '7') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case7:
		return &registers[7];
					goto strswitch0_case8;
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '8') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case8:
		return &registers[8];
					goto strswitch0_case9;
				} else {
					goto strswitch0_default;
				}
			} else if (strswitch0_str[1] == '9') {
				if (strswitch0_str[2] == '\0') {
					strswitch0_case9:
		return &registers[9];
					goto strswitch0_case10;
				} else {
					goto strswitch0_default;
				}
			} else {
				goto strswitch0_default;
			}
		} else {
			strswitch0_default:
		return NULL;
		}
	}
	}
}

#define DRGN_ARCHITECTURE_REGISTERS \
	.register_layout = register_layout, \
	.dwarf_regno_to_internal = dwarf_regno_to_internal, \
	.registers = registers, \
	.num_registers = 41, \
	.stack_pointer_regno = DRGN_REGISTER_NUMBER__r1, \
	.register_by_name = register_by_name
