// Copyright (c) Meta Platforms, Inc. and affiliates.
// SPDX-License-Identifier: LGPL-2.1-or-later

// Create external definitions of accessors.
#define DRGN_ACCESSOR_LINKAGE LIBDRGN_PUBLIC extern
#include "util.h"
#include "drgn_internal.h"
