/* Generated by drgndoc.docstrings -H. */

extern const char drgn_DOC[];
extern const char drgn_IntegerLike_DOC[];
extern const char drgn_Path_DOC[];
extern const char drgn_Program_DOC[];
extern const char drgn_Program___new___DOC[];
extern const char drgn_Program_flags_DOC[];
extern const char drgn_Program_platform_DOC[];
extern const char drgn_Program_core_dump_path_DOC[];
extern const char drgn_Program_language_DOC[];
extern const char drgn_Program___getitem___DOC[];
extern const char drgn_Program___contains___DOC[];
extern const char drgn_Program_variable_DOC[];
extern const char drgn_Program_constant_DOC[];
extern const char drgn_Program_function_DOC[];
extern const char drgn_Program_object_DOC[];
extern const char drgn_Program_symbol_DOC[];
extern const char drgn_Program_symbols_DOC[];
extern const char drgn_Program_stack_trace_DOC[];
extern const char drgn_Program_stack_trace_from_pcs_DOC[];
extern const char drgn_Program_source_location_DOC[];
extern const char drgn_Program_type_DOC[];
extern const char drgn_Program_threads_DOC[];
extern const char drgn_Program_thread_DOC[];
extern const char drgn_Program_main_thread_DOC[];
extern const char drgn_Program_crashed_thread_DOC[];
extern const char drgn_Program_read_DOC[];
extern const char drgn_Program_read_c_string_DOC[];
extern const char drgn_Program_read_u8_DOC[];
extern const char drgn_Program_read_u16_DOC[];
extern const char drgn_Program_read_u32_DOC[];
extern const char drgn_Program_read_u64_DOC[];
extern const char drgn_Program_read_word_DOC[];
extern const char drgn_Program_search_memory_DOC[];
extern const char drgn_Program_search_memory_u16_DOC[];
extern const char drgn_Program_search_memory_u32_DOC[];
extern const char drgn_Program_search_memory_u64_DOC[];
extern const char drgn_Program_search_memory_word_DOC[];
extern const char drgn_Program_search_memory_regex_DOC[];
extern const char drgn_Program_add_memory_segment_DOC[];
extern const char drgn_Program_register_type_finder_DOC[];
extern const char drgn_Program_registered_type_finders_DOC[];
extern const char drgn_Program_set_enabled_type_finders_DOC[];
extern const char drgn_Program_enabled_type_finders_DOC[];
extern const char drgn_Program_register_object_finder_DOC[];
extern const char drgn_Program_registered_object_finders_DOC[];
extern const char drgn_Program_set_enabled_object_finders_DOC[];
extern const char drgn_Program_enabled_object_finders_DOC[];
extern const char drgn_Program_register_symbol_finder_DOC[];
extern const char drgn_Program_registered_symbol_finders_DOC[];
extern const char drgn_Program_set_enabled_symbol_finders_DOC[];
extern const char drgn_Program_enabled_symbol_finders_DOC[];
extern const char drgn_Program_add_type_finder_DOC[];
extern const char drgn_Program_add_object_finder_DOC[];
extern const char drgn_Program_set_core_dump_DOC[];
extern const char drgn_Program_set_kernel_DOC[];
extern const char drgn_Program_set_linux_kernel_custom_DOC[];
extern const char drgn_Program_set_pid_DOC[];
extern const char drgn_Program_set_qemu_qmp_DOC[];
extern const char drgn_Program_modules_DOC[];
extern const char drgn_Program_loaded_modules_DOC[];
extern const char drgn_Program_create_loaded_modules_DOC[];
extern const char drgn_Program_main_module_DOC[];
extern const char drgn_Program_shared_library_module_DOC[];
extern const char drgn_Program_vdso_module_DOC[];
extern const char drgn_Program_relocatable_module_DOC[];
extern const char drgn_Program_linux_kernel_loadable_module_DOC[];
extern const char drgn_Program_extra_module_DOC[];
extern const char drgn_Program_module_DOC[];
extern const char drgn_Program_register_debug_info_finder_DOC[];
extern const char drgn_Program_registered_debug_info_finders_DOC[];
extern const char drgn_Program_set_enabled_debug_info_finders_DOC[];
extern const char drgn_Program_enabled_debug_info_finders_DOC[];
extern const char drgn_Program_debug_info_options_DOC[];
extern const char drgn_Program_load_debug_info_DOC[];
extern const char drgn_Program_load_default_debug_info_DOC[];
extern const char drgn_Program_load_module_debug_info_DOC[];
extern const char drgn_Program_find_standard_debug_info_DOC[];
extern const char drgn_Program_address_size_DOC[];
extern const char drgn_Program_cache_DOC[];
extern const char drgn_Program_config_DOC[];
extern const char drgn_Program_void_type_DOC[];
extern const char drgn_Program_int_type_DOC[];
extern const char drgn_Program_bool_type_DOC[];
extern const char drgn_Program_float_type_DOC[];
extern const char drgn_Program_struct_type_DOC[];
extern const char drgn_Program_union_type_DOC[];
extern const char drgn_Program_class_type_DOC[];
extern const char drgn_Program_enum_type_DOC[];
extern const char drgn_Program_typedef_type_DOC[];
extern const char drgn_Program_pointer_type_DOC[];
extern const char drgn_Program_array_type_DOC[];
extern const char drgn_Program_function_type_DOC[];
extern const char drgn_ProgramFlags_DOC[];
extern const char drgn_ProgramFlags_IS_LINUX_KERNEL_DOC[];
extern const char drgn_ProgramFlags_IS_LIVE_DOC[];
extern const char drgn_ProgramFlags_IS_LOCAL_DOC[];
extern const char drgn_FindObjectFlags_DOC[];
extern const char drgn_FindObjectFlags_CONSTANT_DOC[];
extern const char drgn_FindObjectFlags_FUNCTION_DOC[];
extern const char drgn_FindObjectFlags_VARIABLE_DOC[];
extern const char drgn_FindObjectFlags_ANY_DOC[];
extern const char drgn_ObjectNotFoundError_DOC[];
extern const char drgn_ObjectNotFoundError_name_DOC[];
extern const char drgn_MemorySearchIterator_DOC[];
extern const char drgn_MemorySearchIterator_set_address_range_DOC[];
extern const char drgn_DebugInfoOptions_DOC[];
extern const char drgn_DebugInfoOptions___new___DOC[];
extern const char drgn_DebugInfoOptions_directories_DOC[];
extern const char drgn_DebugInfoOptions_try_module_name_DOC[];
extern const char drgn_DebugInfoOptions_try_build_id_DOC[];
extern const char drgn_DebugInfoOptions_debug_link_directories_DOC[];
extern const char drgn_DebugInfoOptions_try_debug_link_DOC[];
extern const char drgn_DebugInfoOptions_try_procfs_DOC[];
extern const char drgn_DebugInfoOptions_try_embedded_vdso_DOC[];
extern const char drgn_DebugInfoOptions_try_reuse_DOC[];
extern const char drgn_DebugInfoOptions_try_supplementary_DOC[];
extern const char drgn_DebugInfoOptions_kernel_directories_DOC[];
extern const char drgn_DebugInfoOptions_try_kmod_DOC[];
extern const char drgn_KmodSearchMethod_DOC[];
extern const char drgn_KmodSearchMethod_NONE_DOC[];
extern const char drgn_KmodSearchMethod_DEPMOD_DOC[];
extern const char drgn_KmodSearchMethod_WALK_DOC[];
extern const char drgn_KmodSearchMethod_DEPMOD_OR_WALK_DOC[];
extern const char drgn_KmodSearchMethod_DEPMOD_AND_WALK_DOC[];
extern const char drgn_get_default_prog_DOC[];
extern const char drgn_set_default_prog_DOC[];
extern const char drgn_NoDefaultProgramError_DOC[];
extern const char drgn_Module_DOC[];
extern const char drgn_Module_prog_DOC[];
extern const char drgn_Module_name_DOC[];
extern const char drgn_Module_address_ranges_DOC[];
extern const char drgn_Module_address_range_DOC[];
extern const char drgn_Module_build_id_DOC[];
extern const char drgn_Module_object_DOC[];
extern const char drgn_Module_loaded_file_status_DOC[];
extern const char drgn_Module_loaded_file_path_DOC[];
extern const char drgn_Module_loaded_file_bias_DOC[];
extern const char drgn_Module_debug_file_status_DOC[];
extern const char drgn_Module_debug_file_path_DOC[];
extern const char drgn_Module_debug_file_bias_DOC[];
extern const char drgn_Module_supplementary_debug_file_kind_DOC[];
extern const char drgn_Module_supplementary_debug_file_path_DOC[];
extern const char drgn_Module_wants_loaded_file_DOC[];
extern const char drgn_Module_wants_debug_file_DOC[];
extern const char drgn_Module_wanted_supplementary_debug_file_DOC[];
extern const char drgn_Module_try_file_DOC[];
extern const char drgn_MainModule_DOC[];
extern const char drgn_SharedLibraryModule_DOC[];
extern const char drgn_SharedLibraryModule_dynamic_address_DOC[];
extern const char drgn_VdsoModule_DOC[];
extern const char drgn_VdsoModule_dynamic_address_DOC[];
extern const char drgn_RelocatableModule_DOC[];
extern const char drgn_RelocatableModule_address_DOC[];
extern const char drgn_RelocatableModule_section_addresses_DOC[];
extern const char drgn_ExtraModule_DOC[];
extern const char drgn_ExtraModule_id_DOC[];
extern const char drgn_ModuleFileStatus_DOC[];
extern const char drgn_ModuleFileStatus_WANT_DOC[];
extern const char drgn_ModuleFileStatus_HAVE_DOC[];
extern const char drgn_ModuleFileStatus_DONT_WANT_DOC[];
extern const char drgn_ModuleFileStatus_DONT_NEED_DOC[];
extern const char drgn_ModuleFileStatus_WANT_SUPPLEMENTARY_DOC[];
extern const char drgn_WantedSupplementaryFile_DOC[];
extern const char drgn_WantedSupplementaryFile_kind_DOC[];
extern const char drgn_WantedSupplementaryFile_path_DOC[];
extern const char drgn_WantedSupplementaryFile_supplementary_path_DOC[];
extern const char drgn_WantedSupplementaryFile_checksum_DOC[];
extern const char drgn_SupplementaryFileKind_DOC[];
extern const char drgn_SupplementaryFileKind_GNU_DEBUGALTLINK_DOC[];
extern const char drgn_Thread_DOC[];
extern const char drgn_Thread_tid_DOC[];
extern const char drgn_Thread_name_DOC[];
extern const char drgn_Thread_object_DOC[];
extern const char drgn_Thread_stack_trace_DOC[];
extern const char drgn_filename_matches_DOC[];
extern const char drgn_program_from_core_dump_DOC[];
extern const char drgn_program_from_kernel_DOC[];
extern const char drgn_program_from_pid_DOC[];
extern const char drgn_Platform_DOC[];
extern const char drgn_Platform___new___DOC[];
extern const char drgn_Platform_arch_DOC[];
extern const char drgn_Platform_flags_DOC[];
extern const char drgn_Platform_registers_DOC[];
extern const char drgn_Architecture_DOC[];
extern const char drgn_Architecture_X86_64_DOC[];
extern const char drgn_Architecture_I386_DOC[];
extern const char drgn_Architecture_AARCH64_DOC[];
extern const char drgn_Architecture_ARM_DOC[];
extern const char drgn_Architecture_PPC64_DOC[];
extern const char drgn_Architecture_RISCV64_DOC[];
extern const char drgn_Architecture_RISCV32_DOC[];
extern const char drgn_Architecture_S390X_DOC[];
extern const char drgn_Architecture_S390_DOC[];
extern const char drgn_Architecture_UNKNOWN_DOC[];
extern const char drgn_PlatformFlags_DOC[];
extern const char drgn_PlatformFlags_IS_64_BIT_DOC[];
extern const char drgn_PlatformFlags_IS_LITTLE_ENDIAN_DOC[];
extern const char drgn_Register_DOC[];
extern const char drgn_Register_names_DOC[];
extern const char drgn_host_platform_DOC[];
extern const char drgn_Language_DOC[];
extern const char drgn_Language_name_DOC[];
extern const char drgn_Language_C_DOC[];
extern const char drgn_Language_CPP_DOC[];
extern const char drgn_Object_DOC[];
extern const char drgn_Object___new___DOC[];
extern const char drgn_Object_prog__DOC[];
extern const char drgn_Object_type__DOC[];
extern const char drgn_Object_address__DOC[];
extern const char drgn_Object_absent__DOC[];
extern const char drgn_Object_absence_reason__DOC[];
extern const char drgn_Object_bit_offset__DOC[];
extern const char drgn_Object_bit_field_size__DOC[];
extern const char drgn_Object___getattr___DOC[];
extern const char drgn_Object___getitem___DOC[];
extern const char drgn_Object___len___DOC[];
extern const char drgn_Object_value__DOC[];
extern const char drgn_Object_string__DOC[];
extern const char drgn_Object_member__DOC[];
extern const char drgn_Object_subobject__DOC[];
extern const char drgn_Object_address_of__DOC[];
extern const char drgn_Object_read__DOC[];
extern const char drgn_Object_to_bytes__DOC[];
extern const char drgn_Object_from_bytes__DOC[];
extern const char drgn_Object_format__DOC[];
extern const char drgn_AbsenceReason_DOC[];
extern const char drgn_AbsenceReason_OTHER_DOC[];
extern const char drgn_AbsenceReason_OPTIMIZED_OUT_DOC[];
extern const char drgn_AbsenceReason_NOT_IMPLEMENTED_DOC[];
extern const char drgn_NULL_DOC[];
extern const char drgn_cast_DOC[];
extern const char drgn_implicit_convert_DOC[];
extern const char drgn_reinterpret_DOC[];
extern const char drgn_container_of_DOC[];
extern const char drgn_Symbol_DOC[];
extern const char drgn_Symbol___new___DOC[];
extern const char drgn_Symbol_name_DOC[];
extern const char drgn_Symbol_address_DOC[];
extern const char drgn_Symbol_size_DOC[];
extern const char drgn_Symbol_binding_DOC[];
extern const char drgn_Symbol_kind_DOC[];
extern const char drgn_SymbolIndex_DOC[];
extern const char drgn_SymbolIndex___new___DOC[];
extern const char drgn_SymbolIndex___call___DOC[];
extern const char drgn_SymbolBinding_DOC[];
extern const char drgn_SymbolBinding_UNKNOWN_DOC[];
extern const char drgn_SymbolBinding_LOCAL_DOC[];
extern const char drgn_SymbolBinding_GLOBAL_DOC[];
extern const char drgn_SymbolBinding_WEAK_DOC[];
extern const char drgn_SymbolBinding_UNIQUE_DOC[];
extern const char drgn_SymbolKind_DOC[];
extern const char drgn_SymbolKind_UNKNOWN_DOC[];
extern const char drgn_SymbolKind_OBJECT_DOC[];
extern const char drgn_SymbolKind_FUNC_DOC[];
extern const char drgn_SymbolKind_SECTION_DOC[];
extern const char drgn_SymbolKind_FILE_DOC[];
extern const char drgn_SymbolKind_COMMON_DOC[];
extern const char drgn_SymbolKind_TLS_DOC[];
extern const char drgn_SymbolKind_IFUNC_DOC[];
extern const char drgn_StackTrace_DOC[];
extern const char drgn_StackTrace_prog_DOC[];
extern const char drgn_StackFrame_DOC[];
extern const char drgn_StackFrame_name_DOC[];
extern const char drgn_StackFrame_function_name_DOC[];
extern const char drgn_StackFrame_is_inline_DOC[];
extern const char drgn_StackFrame_interrupted_DOC[];
extern const char drgn_StackFrame_pc_DOC[];
extern const char drgn_StackFrame_sp_DOC[];
extern const char drgn_StackFrame___getitem___DOC[];
extern const char drgn_StackFrame___contains___DOC[];
extern const char drgn_StackFrame_locals_DOC[];
extern const char drgn_StackFrame_source_DOC[];
extern const char drgn_StackFrame_symbol_DOC[];
extern const char drgn_StackFrame_register_DOC[];
extern const char drgn_StackFrame_registers_DOC[];
extern const char drgn_SourceLocation_DOC[];
extern const char drgn_SourceLocation_filename_DOC[];
extern const char drgn_SourceLocation_line_DOC[];
extern const char drgn_SourceLocation_column_DOC[];
extern const char drgn_SourceLocation_name_DOC[];
extern const char drgn_SourceLocationList_DOC[];
extern const char drgn_SourceLocationList_prog_DOC[];
extern const char drgn_Type_DOC[];
extern const char drgn_Type_prog_DOC[];
extern const char drgn_Type_kind_DOC[];
extern const char drgn_Type_primitive_DOC[];
extern const char drgn_Type_qualifiers_DOC[];
extern const char drgn_Type_language_DOC[];
extern const char drgn_Type_name_DOC[];
extern const char drgn_Type_tag_DOC[];
extern const char drgn_Type_size_DOC[];
extern const char drgn_Type_length_DOC[];
extern const char drgn_Type_is_signed_DOC[];
extern const char drgn_Type_byteorder_DOC[];
extern const char drgn_Type_type_DOC[];
extern const char drgn_Type_members_DOC[];
extern const char drgn_Type_enumerators_DOC[];
extern const char drgn_Type_parameters_DOC[];
extern const char drgn_Type_is_variadic_DOC[];
extern const char drgn_Type_template_parameters_DOC[];
extern const char drgn_Type_type_name_DOC[];
extern const char drgn_Type_variable_declaration_DOC[];
extern const char drgn_Type_is_complete_DOC[];
extern const char drgn_Type_qualified_DOC[];
extern const char drgn_Type_unqualified_DOC[];
extern const char drgn_Type_unaliased_DOC[];
extern const char drgn_Type_unaliased_kind_DOC[];
extern const char drgn_Type_member_DOC[];
extern const char drgn_Type_has_member_DOC[];
extern const char drgn_TypeMember_DOC[];
extern const char drgn_TypeMember___new___DOC[];
extern const char drgn_TypeMember_object_DOC[];
extern const char drgn_TypeMember_type_DOC[];
extern const char drgn_TypeMember_name_DOC[];
extern const char drgn_TypeMember_bit_offset_DOC[];
extern const char drgn_TypeMember_offset_DOC[];
extern const char drgn_TypeMember_bit_field_size_DOC[];
extern const char drgn_TypeEnumerator_DOC[];
extern const char drgn_TypeEnumerator___new___DOC[];
extern const char drgn_TypeEnumerator_name_DOC[];
extern const char drgn_TypeEnumerator_value_DOC[];
extern const char drgn_TypeParameter_DOC[];
extern const char drgn_TypeParameter___new___DOC[];
extern const char drgn_TypeParameter_default_argument_DOC[];
extern const char drgn_TypeParameter_type_DOC[];
extern const char drgn_TypeParameter_name_DOC[];
extern const char drgn_TypeTemplateParameter_DOC[];
extern const char drgn_TypeTemplateParameter___new___DOC[];
extern const char drgn_TypeTemplateParameter_argument_DOC[];
extern const char drgn_TypeTemplateParameter_name_DOC[];
extern const char drgn_TypeTemplateParameter_is_default_DOC[];
extern const char drgn_TypeKind_DOC[];
extern const char drgn_TypeKind_VOID_DOC[];
extern const char drgn_TypeKind_INT_DOC[];
extern const char drgn_TypeKind_BOOL_DOC[];
extern const char drgn_TypeKind_FLOAT_DOC[];
extern const char drgn_TypeKind_STRUCT_DOC[];
extern const char drgn_TypeKind_UNION_DOC[];
extern const char drgn_TypeKind_CLASS_DOC[];
extern const char drgn_TypeKind_ENUM_DOC[];
extern const char drgn_TypeKind_TYPEDEF_DOC[];
extern const char drgn_TypeKind_POINTER_DOC[];
extern const char drgn_TypeKind_ARRAY_DOC[];
extern const char drgn_TypeKind_FUNCTION_DOC[];
extern const char drgn_TypeKindSet_DOC[];
extern const char drgn_PrimitiveType_DOC[];
extern const char drgn_PrimitiveType_C_VOID_DOC[];
extern const char drgn_PrimitiveType_C_CHAR_DOC[];
extern const char drgn_PrimitiveType_C_SIGNED_CHAR_DOC[];
extern const char drgn_PrimitiveType_C_UNSIGNED_CHAR_DOC[];
extern const char drgn_PrimitiveType_C_SHORT_DOC[];
extern const char drgn_PrimitiveType_C_UNSIGNED_SHORT_DOC[];
extern const char drgn_PrimitiveType_C_INT_DOC[];
extern const char drgn_PrimitiveType_C_UNSIGNED_INT_DOC[];
extern const char drgn_PrimitiveType_C_LONG_DOC[];
extern const char drgn_PrimitiveType_C_UNSIGNED_LONG_DOC[];
extern const char drgn_PrimitiveType_C_LONG_LONG_DOC[];
extern const char drgn_PrimitiveType_C_UNSIGNED_LONG_LONG_DOC[];
extern const char drgn_PrimitiveType_C_BOOL_DOC[];
extern const char drgn_PrimitiveType_C_FLOAT_DOC[];
extern const char drgn_PrimitiveType_C_DOUBLE_DOC[];
extern const char drgn_PrimitiveType_C_LONG_DOUBLE_DOC[];
extern const char drgn_PrimitiveType_C_SIZE_T_DOC[];
extern const char drgn_PrimitiveType_C_PTRDIFF_T_DOC[];
extern const char drgn_Qualifiers_DOC[];
extern const char drgn_Qualifiers_NONE_DOC[];
extern const char drgn_Qualifiers_CONST_DOC[];
extern const char drgn_Qualifiers_VOLATILE_DOC[];
extern const char drgn_Qualifiers_RESTRICT_DOC[];
extern const char drgn_Qualifiers_ATOMIC_DOC[];
extern const char drgn_sizeof_DOC[];
extern const char drgn_alignof_DOC[];
extern const char drgn_offsetof_DOC[];
extern const char drgn_FaultError_DOC[];
extern const char drgn_FaultError_message_DOC[];
extern const char drgn_FaultError_address_DOC[];
extern const char drgn_MissingDebugInfoError_DOC[];
extern const char drgn_ObjectAbsentError_DOC[];
extern const char drgn_OutOfBoundsError_DOC[];
extern const char drgn__linux_helper_per_cpu_ptr_DOC[];
extern const char drgn__linux_helper_task_thread_info_DOC[];
extern const char drgn__linux_helper_task_cpu_DOC[];
extern const char drgn__linux_helper_task_on_cpu_DOC[];
extern const char drgn__linux_helper_pid_task_DOC[];
